unit DllInjector;

interface

uses Windows, SysUtils, UnicodeUtils, DllShare;

type
  EInject = class (Exception)
  end;

  PLoaderData = ^TLoaderData;
  TLoaderData = record
    _LoadLibraryW: function (lpLibFileName: PWideChar): HMODULE; stdcall;
    _GetProcAddress: function (hModule: HMODULE; lpProcName: LPCSTR): FARPROC; stdcall;
    _ExitThread: procedure (dwExitCode: DWORD); stdcall;

    DllFile: array[0..MAX_PATH] of WideChar;
    DllInitProc: array[0..20] of Char;
    DllInitDataSize: DWord;

    { DllInitData follows. }
  end;

  TDllInjector = class
  protected
    FExeFile: WideString;
    FDllFile: WideString;

    FDllInitData: Pointer;
    FDllInitDataSize: DWord;

    FProcInfo: TProcessInformation;
    FProcess: THandle;
    FMainThread: THandle;

    FLdrDataPtr: Pointer;

    procedure InjectIntoProcess;
      function PrepareLoaderDataFor(Process: THandle; DllFile: WideString; const DllInitProc: String;
          DllInitData: Pointer; DllInitDataSize: DWord): TLoaderData;
    procedure InjectLoaderAndRun;
  public
    FDllInitProcName: String;

    constructor Create;

    procedure InjectInto(ExeFile, DllFile: WideString; const DllInitData; DllInitDataSize: DWord);
      procedure Inject;
  end;

procedure InjectDllInto(ExeFile, DllFile: WideString; const DllInitData; DllInitDataSize: DWord);

const
  InjectionFailedExitCode = $FEDA;

procedure LoaderProc(LoaderData: PLoaderData); stdcall;
  procedure LoaderProcEnd;

implementation

procedure LoaderProc(LoaderData: PLoaderData); stdcall;
var
  DLL: HModule;
  InitProc: TDllInitProc;
  DllInitData: Pointer;
begin
  with LoaderData^ do
  begin                          
    DLL := _LoadLibraryW( @DllFile[0] );
    @InitProc := _GetProcAddress(DLL, @DllInitProc[0]);

    if Assigned(InitProc) then
    begin
      DllInitData := Pointer( DWord(LoaderData) + SizeOf(LoaderData^) );
      InitProc(DllInitData, DllInitDataSize);
    end;

    _ExitThread(0);
  end;
end;

  procedure LoaderProcEnd;
  begin
  end;

procedure InjectDllInto(ExeFile, DllFile: WideString; const DllInitData; DllInitDataSize: DWord);
var
  Injector: TDllInjector;
begin
  Injector := TDllInjector.Create;
  try
    Injector.InjectInto(ExeFile, DllFile, DllInitData, DllInitDataSize);
  finally
    Injector.Free;
  end;
end;

{ TDllInjector }

constructor TDllInjector.Create;
begin
  FDllInitProcName := 'InitAfterInjecttion';
end;

procedure TDllInjector.InjectInto(ExeFile, DllFile: WideString; const DllInitData; DllInitDataSize: DWord);
begin
  FExeFile := ExeFile;
  FDllFile := DllFile;

  FDllInitData := @DllInitData;
  FDllInitDataSize := DllInitDataSize;

  Inject;
end;

procedure TDllInjector.Inject;
var
  StartupInfo: TStartupInfo;
begin
  if not FileExists(FExeFile) then
    raise EInject.CreateFmt('Program to run %s doesn''t exist.', [BaseName(FExeFile)]);

  ZeroMemory(@StartupInfo, SizeOf(StartupInfo));
  CreateProcessW(PWideChar(FExeFile), '', NIL, NIL, False, CREATE_SUSPENDED, NIL, NIL, StartupInfo, FProcInfo);

    FMainThread := FProcInfo.hThread;

    FProcess := OpenProcess(PROCESS_QUERY_INFORMATION or PROCESS_CREATE_THREAD or PROCESS_VM_OPERATION
                            or PROCESS_VM_WRITE or PROCESS_TERMINATE, False, FProcInfo.dwProcessId);
    if FProcess = 0 then
      raise EInject.CreateFmt('Cannot open the process for %s.', [BaseName(FExeFile)]);

  try
    try
      InjectIntoProcess;
    except
      TerminateProcess(FProcess, InjectionFailedExitCode);
      raise;
    end;
  finally
    CloseHandle(FProcess);
  end;
end;

procedure TDllInjector.InjectIntoProcess;
var
  LoaderData: TLoaderData;
  LdrDataSize, BytesWritten: DWord;
  DllInitDataPtr: Pointer;
begin
  LoaderData := PrepareLoaderDataFor(FProcess, FDllFile, FDllInitProcName, FDllInitData, FDllInitDataSize);

    LdrDataSize := SizeOf(TLoaderData) + FDllInitDataSize;

  FLdrDataPtr := VirtualAllocEx(FProcess, NIL, LdrDataSize, MEM_COMMIT, PAGE_READWRITE);
  try
      DllInitDataPtr := Pointer( DWord(FLdrDataPtr) + SizeOf(TLoaderData) );

    if    not WriteProcessMemory(FProcess, FLdrDataPtr, @LoaderData, SizeOf(TLoaderData), BytesWritten)
       or not WriteProcessMemory(FProcess, DllInitDataPtr, FDllInitData, FDllInitDataSize, BytesWritten) then
      raise EInject.Create('Coudln''t write loader data into virtual-allocated memory of new process.');

    InjectLoaderAndRun;
  finally
    VirtualFreeEx(FProcess, FLdrDataPtr, LdrDataSize, MEM_DECOMMIT);
  end;
end;

  function TDllInjector.PrepareLoaderDataFor(Process: THandle; DllFile: WideString; const DllInitProc: String;
      DllInitData: Pointer; DllInitDataSize: DWord): TLoaderData;
  var
    Data: TLoaderData;
    Kernel: HModule;
  begin
    ZeroMemory(@Data, SizeOf(Data));

    Kernel := GetModuleHandle(Kernel32);
      Data._LoadLibraryW := GetProcAddress(Kernel, 'LoadLibraryW');
      Data._GetProcAddress := GetProcAddress(Kernel, 'GetProcAddress');
      Data._ExitThread := GetProcAddress(Kernel, 'ExitThread');

    DllFile := ExpandFileName(DllFile);
    if not FileExists(DllFile) then
      raise EInject.CreateFmt('Library to be loaded %s doesn''t exist.', [BaseName(DllFile)]);
    if Length(DllFile) > Length(Data.DllFile) - 1 then
      raise EInject.CreateFmt('Too long absolute path of DLL to be loaded:'#10'%s', [DllFile]);

      Move(DllFile[1], Data.DllFile[0], Length(DllFile) * 2);

    // 1 extra byte is needed for #0 char since the field is used as PChar in LoaderProc.
    if Length(DllInitProc) > Length(Data.DllInitProc) - 1 then
      raise EInject.CreateFmt('Too long name of to be loaded DLL''s init procedure: %s', [DllInitProc]);

      Move(DllInitProc[1], Data.DllInitProc[0], Length(DllInitProc));

    Data.DllInitDataSize := DllInitDataSize;

    Result := Data;
  end;

procedure TDllInjector.InjectLoaderAndRun;
var
  LdrProcLength, BytesWritten, ThreadID: DWord;
  LdrCodePtr: Pointer;
  Thread: THandle;
begin
    LdrProcLength := DWord(@LoaderProcEnd) - DWord(@LoaderProc);

  LdrCodePtr := VirtualAllocEx(FProcess, NIL, LdrProcLength, MEM_COMMIT, PAGE_READWRITE);
  if not WriteProcessMemory(FProcess, LdrCodePtr, @LoaderProc, LdrProcLength, BytesWritten) then
    raise EInject.Create('Coudln''t write loader code into virtual-allocated memory of new process.');

  try
    Thread := CreateRemoteThread(FProcess, NIL, 0, LdrCodePtr, FLdrDataPtr, 0, ThreadID);
    if Thread = 0 then
      raise EInject.Create('Coudln''t create loader''s remote thread.');

    try
      WaitForSingleObject(Thread, INFINITE);

      if ResumeThread(FMainThread) = $FFFFFFFF then
        raise EInject.Create('Coudln''t resume main program''s thread (after execution of loader''s thread).');
    finally
      CloseHandle(FMainThread);
    end;
  finally
    VirtualFreeEx(FProcess, LdrCodePtr, LdrProcLength, MEM_DECOMMIT);
  end;
end;

end.
