unit ExeUtils;

interface

uses Windows, SysUtils;

type
  EExeUtils = class (Exception) end;
    EHookImport = class (EExeUtils) end;
    EWriteProcessMemory = class (EExeUtils) end;

  PImageImportDescriptor = ^TImageImportDescriptor;
  TImageImportDescriptor = packed record
    OriginalFirstThunk: DWord;
    TimeDateStamp: DWord;
    ForwarderChain: DWord;
    Name: DWord;
    FirstThunk: DWord;
  end;

  TJmpInstruction = packed record
    JmpOpcode: Byte;
    Distance:  DWord;
  end;

function HookImport(const ExportingDllName: WideString; const ProcName: String;
    NewProcAddress: Pointer): Boolean;
function ProcessFileName: WideString;
function OpenSelfProcess: THandle;
function WriteProcessMemory(Process: THandle; Address: Pointer; const Buffer; Size: DWord;
    const ErrorMsg: String = 'Patching'): DWord;

function IncPointer(Ptr: Pointer; By: Integer): Pointer;     
// Generates only near jumps.
function MakeJMP(JmpInstrAddr, JumpDestAddr: Pointer): TJmpInstruction;

implementation

// I got the idea for this func somewhere in the Internet a few years ago but I don't remember from where.
function HookImport(const ExportingDllName: WideString; const ProcName: String; NewProcAddress: Pointer): Boolean;
var
  ImageBase, OldProtect: Cardinal;
  PEHeader: PImageNtHeaders;
  PImport: PImageImportDescriptor;
  PRVA_Import: LPDWORD;
  ProcAddress: Pointer;
begin
  Result := False;

  ImageBase := GetModuleHandle(NIL);
  PEHeader := Pointer(Int64(ImageBase) + PImageDosHeader(ImageBase)._lfanew);
  PImport := Pointer(PEHeader.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress + ImageBase);

  ProcAddress := GetProcAddress( GetModuleHandleW( PWideChar(ExportingDllName) ), PChar(ProcName) );
  if ProcAddress = NIL then
    raise EHookImport.CreateFmt('Cannot find exported procedure %s in DLL %s to hook.',
                                [ProcName, ExportingDllName]);

  while PImport.Name <> 0 do
  begin
    PRVA_Import := LPDWORD(pImport.FirstThunk + ImageBase);

    while PRVA_Import^ <> 0 do
    begin
      if PPointer(PRVA_Import)^ = ProcAddress then
      begin
        Result := True;

        if not VirtualProtect(PPointer(PRVA_Import), SizeOf(Pointer), PAGE_READWRITE, OldProtect) then
          raise EHookImport.CreateFmt('Cannot VirtualProtect import address of %s.', [ProcName]);

        try
          PPointer(PRVA_Import)^ := NewProcAddress;
        finally
          VirtualProtect(PPointer(PRVA_Import), SizeOf(Pointer), OldProtect, OldProtect);
        end;
      end;

      Inc(PRVA_Import);
    end;
    Inc(PImport);
  end;
end;

function ProcessFileName: WideString;
begin
  SetLength( Result, MAX_PATH );
  SetLength( Result, GetModuleFileNameW(0, @Result[1], MAX_PATH) );
end;

function OpenSelfProcess: THandle;
begin
  Result := OpenProcess(PROCESS_QUERY_INFORMATION or PROCESS_CREATE_THREAD or PROCESS_VM_OPERATION
                        or PROCESS_VM_READ or PROCESS_VM_WRITE or PROCESS_TERMINATE, False, GetCurrentProcessId);
end;

function WriteProcessMemory(Process: THandle; Address: Pointer; const Buffer; Size: DWord;
    const ErrorMsg: String = 'Patching'): DWord;
begin
  if not Windows.WriteProcessMemory(Process, Address, @Buffer, Size, Result) then
    raise EWriteProcessMemory.CreateFmt('%s - cannot write to process mem at %.8x. GetLastError = %d [%s].',
                                        [ErrorMsg, DWord(Address), GetLastError, SysErrorMessage(GetLastError)]);
end;

function IncPointer(Ptr: Pointer; By: Integer): Pointer;
begin
  Result := Pointer( DWord(Ptr) + By );
end;

function MakeJMP(JmpInstrAddr, JumpDestAddr: Pointer): TJmpInstruction;
const
  NearJmpOpcode: Byte = $E9;
  NearJmpInstrLength = SizeOf(TJmpInstruction);
var
  Distance: Integer;
begin
  Distance := DWord(JumpDestAddr) - (DWord(JmpInstrAddr) + SizeOf(TJmpInstruction));
  if Distance < 0 then
    Distance := not Distance + 1;

  Result.JmpOpcode := NearJmpOpcode;
  Result.Distance := Distance;
end;

end.
