library translation;

uses
  Windows,
  SysUtils,
  MMSystem,
  StrConv,
  DllShare,
  ExeUtils,
  TrueWorldDLL;

{$WARN SYMBOL_PLATFORM OFF}

const
  GameExeVerificationFailed = 'The game''s EXE differs from original:'#10#10 +
                              '    %s'#10#10 +
                              'You may have wrong game version. Continue?'#10#10 +
                              '    Yes - ignore this inconsistency and attempt to patch the game nevertheless.'#10 +
                              '    No - quit.'#10#10 +
                              'Note that checks can be disabled via loader''s (Translation.exe) command line options.';

  OtherErrorMsg             = 'Sorry, the translation library has encountered an error:'#10#10 +
                              '    %s'#10#10 +
                              'Ignore and continue?';

var
  Settings: TGameTranslationDllInitData;
  WasAlreadyInit: Integer = 0;
  MainWindowHandle: HWND;
  GameCompletionMessageDisplayed: Boolean;

{$I IncludeDLL.pas}

procedure EndProcessOnError;
begin
  ExitProcess($AFFE);
end;

function MessageBox(Text: WideString; Flags: DWord = 0): DWord;
begin
  if MainWindowHandle = 0 then
    Flags := Flags or mb_SystemModal;
  Result := MessageBoxW(MainWindowHandle, PWideChar(Text), PWideChar(AppWindowTitle), Flags);
end;

procedure HandleException(E: Exception; Msg: WideString = '');
begin
  if Msg = '' then
    Msg := OtherErrorMsg;

  if MessageBox(WideFormat(Msg, [E.Message]), mb_IconWarning or mb_YesNo or mb_DefButton2) = idNo then
    EndProcessOnError;
end;

{ Patch functions }

function HookedGetGlyphOutline(DC: HDC; uChar, uFormat: UINT;
    const lpgm: TGlyphMetrics; cbBuffer: DWORD; lpvBuffer: Pointer; const lpmat2: TMat2): DWORD; stdcall;
begin
  Result := GetGlyphOutlineW(DC, DWord(SJISCharToWide(uChar)), uFormat, lpgm, cbBuffer, lpvBuffer, lpmat2);
end;

function HookedCreateFont(nHeight, nWidth, nEscapement, nOrientaion, fnWeight: Integer;
    fdwItalic, fdwUnderline, fdwStrikeOut, fdwCharSet, fdwOutputPrecision,
    fdwClipPrecision, fdwQuality, fdwPitchAndFamily: DWORD; lpszFace: PAnsiChar): HFONT; stdcall;
const
  MaxHeight = $36;
var
  FontName: PWideChar;
begin
  FontName := PWideChar( WideString(Settings.FontName) );

  // Only change text window font's width.
  if nWidth = $1B then
  begin
    nWidth := Settings.FontWidth;
    nHeight := Settings.FontHeight;
    // The game engine has hardcoded font sizes and line heights.
    if nHeight > MaxHeight then
      HandleException(EAbort.CreateFmt('Font height is too large (%d > %d).', [nHeight, MaxHeight]));
  end;

  Result := CreateFontW(nHeight, nWidth, nEscapement, nOrientaion, Settings.FontWeight,
                        fdwItalic, fdwUnderline, fdwStrikeOut, ANSI_CHARSET, fdwOutputPrecision,
                        fdwClipPrecision, fdwQuality, fdwPitchAndFamily, FontName);
end;

function HookedCreateWindowEx(dwExStyle: DWORD; lpClassName: PAnsiChar;
    lpWindowName: PAnsiChar; dwStyle: DWORD; X, Y, nWidth, nHeight: Integer;
    hWndParent: HWND; hMenu: HMENU; hInstance: HINST; lpParam: Pointer): HWND; stdcall;
begin
  Result := CreateWindowExW(dwExStyle, PWideChar( WideString(lpClassName) ),
                            PWideChar(AppWindowTitle), dwStyle, X, Y, nWidth, nHeight,
                            hWndParent, hMenu, hInstance, lpParam);
  MainWindowHandle := Result;
end;

function GetTempPath: WideString;
begin
  SetLength(Result, 1024);
  SetLength(Result, GetTempPathW(1024, @Result[1]));
end;
                             
function GetTempFileName(const Prefix: WideString): WideString;
begin
  SetLength(Result, 1024);
  if GetTempFileNameW(PWideChar(GetTempPath), PWideChar(Prefix), 0, @Result[1]) = 0 then
    RaiseLastOSError
    else
      Result := PWideChar(Result);
end;

function HookedCreateFile(lpFileName: PAnsiChar; dwDesiredAccess, dwShareMode: DWORD;
  lpSecurityAttributes: PSecurityAttributes; dwCreationDisposition, dwFlagsAndAttributes: DWORD;
  hTemplateFile: THandle): THandle; stdcall;
var
  FN: String;
  Res: HRSRC;
  Loaded: HGLOBAL;
  Data: Pointer;
  TempPath: WideString;
  Temp: THandle;
  Bytes: DWord;
begin
  if (LowerCase(ExtractFileName(lpFileName)) = 'title_main_r.mes') and not GameCompletionMessageDisplayed then
  begin
    GameCompletionMessageDisplayed := True;
    MessageBox(GameCompletionMessage, mb_IconInformation);
  end;

  // voice.dat starts with a DWORD - a buffer size, which on some ISOs is
  // "aaaa" ($61616161). Due to this the game hangs on start.
  if LowerCase(ExtractFileName(lpFileName)) = 'voice.dat' then
  begin
    Result := INVALID_HANDLE_VALUE;
    Exit;
  end;
  
  if LowerCase(ExtractFileExt(lpFileName)) = '.mes' then
  begin
    FN := LowerCase(ExtractFileName(lpFileName));
    FN := Copy(FN, 1, Length(FN) - 4);
    if ForceReplayTexts and (Copy(FN, Length(FN) - 1, 2) <> '_r') then
      FN := FN + '_r';
    if (ForceFirstText <> '') and (FN = 'cmn_00') then
      FN := ForceFirstText;
    Res := FindResource(HInstance, PChar(FN), RT_RCDATA);
    Temp := INVALID_HANDLE_VALUE;
    OutputDebugString(PChar(Format('File: %s - %d', [FN, Res])));
    if Res <> 0 then
    try 
      Loaded := LoadResource(HInstance, Res);
      Win32Check(Loaded <> 0);
      Data := LockResource(Loaded);
      Win32Check(Data <> NIL);
      TempPath := GetTempFileName(Language[1] + FN[1] + FN[6]);
      if TempPath = '' then
        TempPath := WideFormat('%s%s-%s.%s', [GetTempPath, Language, WideString(FN), 'mes']);
      Temp := CreateFileW(PWideChar(TempPath), FILE_SHARE_DELETE + GENERIC_READ + GENERIC_WRITE,
                          0, NIL, CREATE_ALWAYS, FILE_FLAG_DELETE_ON_CLOSE, 0);
      Win32Check(Temp <> INVALID_HANDLE_VALUE);
      Bytes := 0;
      WriteFile(Temp, Data^, SizeofResource(HInstance, Res), Bytes, NIL);
      Win32Check(SetFilePointer(Temp, 0, NIL, SEEK_SET) <> $FFFFFFFF);
      if Bytes <> SizeofResource(HInstance, Res) then
        raise EOSError.CreateFmt('Cannot write %s (written %d, not %d).', [TempPath, Bytes, SizeofResource(HInstance, Res)]);
      Result := Temp;
      Exit;
    except
      on E: Exception do
      begin
        CloseHandle(Temp);
        HandleException(E, 'Cannot slip in the translated script.');
      end;
    end;
  end;

  Result := CreateFile(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes,
                       dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile);
end;

procedure Init;
begin
  if InterlockedExchange(WasAlreadyInit, 1) <> 0 then
    Exit;

  if not Settings.SkipExeChecks then
    try
      VerifyGameExe;
    except
      on E: Exception do
        HandleException(E, GameExeVerificationFailed);
    end;

  try
    HookImport(Gdi32, 'CreateFontA', @HookedCreateFont);
    HookImport(User32, 'CreateWindowExA', @HookedCreateWindowEx);
    HookImport(Kernel32, 'CreateFileA', @HookedCreateFile);

    PatchSizeOfDrawnText(Settings);
      HookImport(Gdi32, 'GetGlyphOutlineA', @HookedGetGlyphOutline);
  except
    on E: Exception do
      HandleException(E);
  end;
end;

{ Entry point }

procedure InitAfterInjecttion(NewSettings: PGameTranslationDllInitData; DataSize: DWord); stdcall;
begin
  Settings := NewSettings^;
  Init;
end;

procedure InitWithDefaultSettings(uTimerID, uMessage: UINT; dwUser, dw1, dw2: DWORD); stdcall;
begin
  Init;
end;

exports InitAfterInjecttion;

begin
  SetDefaultDllSettingsTo(Settings);

  //   This is useful when the library was injected using some other loader but Translation.exe.
  // 3rd party loaders won't call InitAfterInjecttion.
  //   However, note that it won't hook early calls (e.g. to CreateFontA).
  timeSetEvent(200, 20, InitWithDefaultSettings, 0, TIME_ONESHOT);
end.
