program TrueCake;

{$APPTYPE CONSOLE}
{$WARN SYMBOL_PLATFORM OFF}

//{$DEFINE RELEASE}

{$IFDEF RELEASE}
{$D-}
{$R Encrypted.res}    // run brcc32 on one of Encrypted-*\Encrypted.rc.
{$ENDIF}

uses
  FastMM4,    // Borland's memory manager in D7 is absolutely horrible at multi-threading.
  SysUtils,
  DateUtils,
  Math,
  Windows,
  WinInet,
  CommDlg;

// ------------------

const
  AES_KEY_LENGTH = 32;
  PROV_RSA_AES = 24;
  CRYPT_VERIFYCONTEXT = $F0000000;
  ALG_CLASS_HASH = 4 shl 13;
  ALG_TYPE_ANY = 0;
  ALG_SID_SHA_512 = 14;
  CALG_SHA_512 = ALG_CLASS_HASH + ALG_TYPE_ANY + ALG_SID_SHA_512;
  HP_HASHVAL = 2;
  PLAINTEXTKEYBLOB = 8;
  CUR_BLOB_VERSION = 2;
  ALG_CLASS_DATA_ENCRYPT = 3 shl 13;
  ALG_TYPE_BLOCK = 3 shl 9;
  ALG_SID_AES_256 = 16;
  CALG_AES_256 = ALG_CLASS_DATA_ENCRYPT + ALG_TYPE_BLOCK + ALG_SID_AES_256;
  KP_IV = 1;
  KP_PADDING = 3;
  KP_MODE = 4;
  PKCS5_PADDING = 1;

type
  HCRYPTPROV = Pointer;
  HCRYPTKEY = Pointer;
  HCRYPTHASH = Pointer;

  BLOBHEADER = packed record
    bType: Byte;
    bVersion: Byte;
    reserved: Word;
    aiKeyAlg: DWORD;
  end;

  TPlainTextKeyBlob = packed record
    hdr: BLOBHEADER;
    dwKeySize: DWORD;
    rgbKeyData: array[0..AES_KEY_LENGTH-1] of Byte;
  end;

function CryptAcquireContextA(var phProv: HCRYPTPROV; szContainer: PChar;
  szProvider: PChar; dwProvType: DWORD; dwFlags: DWORD): Boolean; stdcall; external advapi32;

function CryptCreateHash(hProv: HCRYPTPROV; Algid: DWORD; hKey: HCRYPTKEY;
  dwFlags: DWORD; var phHash: HCRYPTHASH): Boolean; stdcall; external advapi32;

function CryptHashData(hHash: HCRYPTHASH; const pbData; dwDataLen: DWORD;
  dwFlags: DWORD): Boolean; stdcall; external advapi32;

function CryptGetHashParam(hHash: HCRYPTHASH; dwParam: DWORD; const pbData;
  var pdwDataLen: DWORD; dwFlags: DWORD): Boolean; stdcall; external advapi32;

function CryptImportKey(hProv: HCRYPTPROV; const pbData; dwDataLen: DWORD;
  hPubKey: HCRYPTKEY; dwFlags: DWORD; var phKey: HCRYPTKEY): Boolean; stdcall; external advapi32;

function CryptSetKeyParam(hKey: HCRYPTKEY; dwParam: DWORD; const pbData;
  dwFlags: DWORD): Boolean; stdcall; external advapi32;

function CryptDecrypt(hKey: HCRYPTKEY; hHash: HCRYPTHASH; Final: Boolean;
  dwFlags: DWORD; var pbData; var pdwDataLen: DWORD): Boolean; stdcall; external advapi32;

function CryptDestroyKey(hKey: HCRYPTKEY): Boolean; stdcall; external advapi32;
function CryptDestroyHash(hHash: HCRYPTHASH): Boolean; stdcall; external advapi32;
function CryptReleaseContext(hProv: HCRYPTPROV; dwFlags: DWORD): Boolean; stdcall; external advapi32;

// ------------------

function CalcHash(hProv: HCRYPTPROV; const Data: String): String;
var
  hHash: HCRYPTHASH;
  Len: DWORD;
begin
  if not CryptCreateHash(hProv, CALG_SHA_512, NIL, 0, hHash) then
    RaiseLastOSError;
  try
    if not CryptHashData(hHash, Data[1], Length(Data), 0) then
      RaiseLastOSError;
    Len := 64;
    SetLength(Result, Len);
    if not CryptGetHashParam(hHash, HP_HASHVAL, Result[1], Len, 0) then
      RaiseLastOSError;
    SetLength(Result, Len);
  finally
    CryptDestroyHash(hHash);
  end;
end;

function Decrypt(hProv: HCRYPTPROV; const EncrKey, EncrIV, Ciphertext: String): String;
const
  Padding: DWORD = PKCS5_PADDING;
var
  Len: DWORD;
  Key: TPlainTextKeyBlob;
  hKey: HCRYPTKEY;
begin
  FillChar(Key, SizeOf(Key), 0);
  Key.hdr.bType := PLAINTEXTKEYBLOB;
  Key.hdr.bVersion := CUR_BLOB_VERSION;
  Key.hdr.aiKeyAlg := CALG_AES_256;
  Key.dwKeySize := AES_KEY_LENGTH;
  Move(EncrKey[1], Key.rgbKeyData, Min(Length(EncrKey), AES_KEY_LENGTH));

  if not CryptImportKey(hProv, Key, SizeOf(Key), NIL, 0, hKey) then
    RaiseLastOSError;
  try
    if Length(EncrIV) < 16 then
      raise EInvalidArgument.CreateFmt('Decrypt: EncrIV must be at least 16 bytes long (%d given).', [Length(EncrIV)]);
    if not CryptSetKeyParam(hKey, KP_IV, EncrIV[1], 0) then
      RaiseLastOSError;
    if not CryptSetKeyParam(hKey, KP_PADDING, Padding, 0) then
      RaiseLastOSError;
    Result := Ciphertext;
    Len := Length(Result);
    if not CryptDecrypt(hKey, NIL, True, 0, Result[1], Len) then
      // Not raising because exceptions are expensive and failing to decrypt is
      // a normal condition for this program.
      Result := ''
    else
      SetLength(Result, Len);
  finally
    CryptDestroyKey(hKey);
  end;
end;

var
  IniFile: array[0..MAX_PATH-1] of Char;

function ReadINI(const Section, Key: String; const Default: String = ''): String;
begin
  SetLength(Result, 255);
  SetLength(Result, GetPrivateProfileString(PChar(Section), PChar(Key), PChar(Default), @Result[1], Length(Result), @IniFile[0]));
end;

procedure WriteINI(const Section, Key, Value: String);
begin
  if not WritePrivateProfileString(PChar(Section), PChar(Key), PChar(Value), @IniFile[0]) then
    RaiseLastOSError;
end;

// From Classes.pas.
procedure BinToHex(Buffer, Text: PChar; BufSize: Integer);
const
  Convert: array[0..15] of Char = '0123456789ABCDEF';
var
  I: Integer;
begin
  for I := 0 to BufSize - 1 do
  begin
    Text[0] := Convert[Byte(Buffer[I]) shr 4];
    Text[1] := Convert[Byte(Buffer[I]) and $F];
    Inc(Text, 2);
  end;
end;

// From Classes.pas.
function HexToBin(Text, Buffer: PChar; BufSize: Integer): Integer;
const
  Convert: array['0'..'f'] of SmallInt =
    ( 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1,
     -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,
     -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
     -1,10,11,12,13,14,15);
var
  I: Integer;
begin
  I := BufSize;
  while I > 0 do
  begin
    if not (Text[0] in ['0'..'f']) or not (Text[1] in ['0'..'f']) then Break;
    Buffer[0] := Char((Convert[Text[0]] shl 4) + Convert[Text[1]]);
    Inc(Buffer);
    Inc(Text, 2);
    Dec(I);
  end;
  Result := BufSize - I;
end;

// This one is unused right now because (as expected though) INI functions are unbearably slow. I'm not feeling like implementing a thread-safe memory index right now...
function CalcHashCaching(hProv: HCRYPTPROV; const Data: String): String;
var
  S: String;
begin
  S := ReadINI('Hashes', Data);
  if (S = '') or (Length(S) mod 2 <> 0) then
  begin
    Result := CalcHash(hProv, Data);
    SetLength(S, Length(Result) * 2);
    BinToHex(@Result[1], @S[1], Length(Result));
    WriteINI('Hashes', Data, S);
  end
  else
  begin
    SetLength(Result, Length(S) div 2);
    HexToBin(@S[1], @Result[1], Length(Result));
  end;
end;

// ------------------

const
  Salt = 'M@RbLeC@KeG@Me';
  WinMark = 'ZomG U WiN! ';

var
  WorkCounter: Int64;
  WorkCounterCriticalSection: TRTLCriticalSection;
  WorkMax: Int64;                     // always not more than around 10^5^3.
  WorkKeys: array of record
    Key: String;                      // 'LineN ArticleN'.
    LineNumber: Integer;
  end;
  EncryptedItems: array of String;
  WonKeys: array[0..499] of record
    Decrypted: String;
    Encrypted: Integer;
    K1, K2, K3: Integer;
    Filled: Integer;
  end;
  WonKeysCounter: Integer;

procedure NewWin(Decrypted: String; EncryptedIndex: Integer; C1, C2, C3: Integer);
var
  WinCount: Integer;
begin
  // Encrypted data may end on a line break (kh, kh, POSIX).
  while Decrypted[Length(Decrypted)] <= ' ' do
    SetLength(Decrypted, Length(Decrypted) - 1);
  WinCount := InterlockedExchangeAdd(WonKeysCounter, 1);
  WonKeys[WinCount].Decrypted := Decrypted;
  WonKeys[WinCount].Encrypted := EncryptedIndex;
  WonKeys[WinCount].K1 := C1;
  WonKeys[WinCount].K2 := C2;
  WonKeys[WinCount].K3 := C3;
  InterlockedIncrement(WonKeys[WinCount].Filled);
end;

procedure ThreadWork(hProv: HCRYPTPROV; C1, C2, C3: Integer);
var
  Hash, Decrypted: String;
  I: Integer;
begin
  Hash := CalcHash(hProv, WorkKeys[C1].Key + ',' + WorkKeys[C2].Key + ',' + WorkKeys[C3].Key + ',' + Salt);
  for I := 0 to Length(EncryptedItems) - 1 do
  begin
    Decrypted := Decrypt(hProv, Copy(Hash, 1, 32), Copy(Hash, 33, 16), EncryptedItems[I]);
    if (Length(Decrypted) > Length(WinMark)) and CompareMem(@Decrypted[1], @WinMark[1], Length(WinMark)) then
      NewWin(Decrypted, I, C1, C2, C3);
  end;
end;

function ThreadProc(Parameter: Pointer): Integer;
var
  hProv: HCRYPTPROV;
  Counter: Int64;
begin
  Result := 0;

  try
    // CRYPT_VERIFYCONTEXT is not required but on XP without this flag it fails
    // with NTE_BAD_KEYSET (0x80090016). On Win10 it works with dwFlags of 0.
    if not CryptAcquireContextA(hProv, NIL, NIL, PROV_RSA_AES, CRYPT_VERIFYCONTEXT) then
      RaiseLastOSError;
    try
      while True do
      begin
        // Request _Atomic!
        EnterCriticalSection(WorkCounterCriticalSection);
        try
          Counter := WorkCounter;
          Inc(WorkCounter);
        finally
          LeaveCriticalSection(WorkCounterCriticalSection);
        end;
        if Counter >= WorkMax then
          Break;
        ThreadWork(hProv,
          Counter mod Length(WorkKeys),
          Counter div Length(WorkKeys) mod Length(WorkKeys),
          Counter div Length(WorkKeys) div Length(WorkKeys) mod Length(WorkKeys));
  //Sleep(100);
      end;
    finally
      CryptReleaseContext(hProv, 0);
    end;
  except
    // try/except not really necessary but FastMM complains if thread exits on Borland's standard exception handler.
    Result := 1;
  end;
end;

function FetchArticleID(const URL: String): Integer;
  function StrPrefixToInt(const S: String; A: Integer = 1): Integer;
  var
    Z: Integer;
  begin
    while S[A] in [' ', #9] do
      Inc(A);
    Z := 0;
    while S[A + Z] in ['0'..'9'] do
      Inc(Z);
    Result := StrToInt(Copy(S, A, Z));
  end;

const
  KYM = '<article class=''entry'' id=''entry_';
  Rulk = '"wgArticleId":';
var
  InetHandle, URLHandle: HInternet;
  Data: String;
  BytesRead: DWORD;
  I: Integer;
begin
  InetHandle := InternetOpen(PChar(''), INTERNET_OPEN_TYPE_PRECONFIG, NIL, NIL, 0);

  try
    URLHandle := InternetOpenURL(InetHandle, PChar(URL), NIL, 0,
      INTERNET_FLAG_EXISTING_CONNECT + INTERNET_FLAG_KEEP_CONNECTION + INTERNET_FLAG_NO_COOKIES + INTERNET_FLAG_NO_UI + INTERNET_FLAG_NO_CACHE_WRITE, 0);

    try
      SetLength(Data, 65536);
      if not InternetReadFile(URLHandle, @Data[1], Length(Data), BytesRead) then
        RaiseLastOSError;
      SetLength(Data, BytesRead);

      I := Pos(KYM, Data);
      if I <> 0 then
      begin
        Result := StrPrefixToInt(Data, I + Length(KYM));
        Exit;
      end;

      I := Pos(Rulk, Data);
      if I <> 0 then
      begin
        Result := StrPrefixToInt(Data, I + Length(Rulk));
        Exit;
      end;

      Result := 0;  // make the compiler happy.
      raise EInvalidArgument.CreateFmt('Can''t find article ID in %s', [URL]);
    finally
      InternetCloseHandle(URLHandle);
    end;
  finally
    InternetCloseHandle(InetHandle);
  end;
end;

function FetchArticleIdCaching(const URL: String): Integer;
begin
  Result := StrToInt(ReadINI('Articles', URL, '0'));
  if Result = 0 then
  begin
    Result := FetchArticleID(URL);
    WriteINI('ArticleIDs', URL, IntToStr(Result));
  end;
end;

function ParseKeyLine(S: String): String;
var
  I: Integer;
  Num: String;
  LineN, ArtN: Integer;
begin
  S := S + ' ';
  Num := '';
  LineN := -1;
  ArtN := -1;

  for I := 1 to Length(S) do
    case S[I] of
    '"',  {UTF-8 BOM} #$EF, #$BB, #$BF:
      Continue;
    ',', ';', ' ', #9:
      begin
        if Num = '' then
          Continue
        else if LineN < 0 then
          LineN := StrToInt(Num)
        else if ArtN < 0 then
        begin
          if Pos('://', Num) = 0 then
            ArtN := StrToInt(Num)
          else
            ArtN := FetchArticleIdCaching(Num);
          // Ignore remaining columns, if any. CakeSheet puts original URL on the 3rd one.
          Break;
        end
        else
          raise EInvalidArgument.CreateFmt('Too many numbers on line: %s', [S]);
        Num := '';
      end;
    else
      Num := Num + S[I];
    end;

  Result := Format('%d %d', [LineN, ArtN]);
end;

// Utils.pas.
procedure CopyToClipboard(Str: WideString);
  procedure CopyToClipUsing(const Data: THandle);
  var
    DataPtr: Pointer;
  begin
    DataPtr := GlobalLock(Data);
    try
      Move(Str[1], DataPtr^, Length(Str) * 2);
    finally
      GlobalUnlock(Data);
    end;

    OpenClipboard(0);
    try
      EmptyClipboard;
      SetClipboardData(CF_UNICODETEXT, Data);
    finally
      CloseClipboard;
    end;
  end;

var
  Data: THandle;
begin
  Str := Str + #0;
  Data := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Length(Str) * 2);
  try
    CopyToClipUsing(Data);
  except
    GlobalFree(Data);
  end;
end;

type
  TMineOptions = record
    KeyFile: String;
    EncryptedResourcesPrefix: String;
    EncryptedPath: String;
    MaxThreads: Integer;
    Wait: Boolean;
  end;

procedure MineForTheWin(const Options: TMineOptions);

  function StringRepeat(const S: String; Times: Integer): String;
  begin
    SetLength(Result, Length(S) * Times);
    for Times := Times downto 1 do
      Move(S[1], Result[Length(S) * Times], Length(S));
  end;

var
  LastWon: Integer;

  function FlushWonKeys(NextUnwon: Integer): Boolean;
  begin
    Result := False;

    while LastWon < NextUnwon do
      with WonKeys[LastWon] do
      begin
        if InterlockedExchangeAdd(Filled, 0) = 0 then
          Break;
        WriteLn(Format('  key# %.3d lines# %.3d+%.3d+%.3d %s', [Encrypted, WorkKeys[K1].LineNumber, WorkKeys[K2].LineNumber, WorkKeys[K3].LineNumber, Decrypted]));
        Inc(LastWon);
        Result := True;
      end;

    if Result then
      WriteLn;
  end;

var
  I, KI: Integer;
  ID: DWORD;
  Threads: array of THandle;
  Res: HRSRC;
  Resource: HGLOBAL;
  Data: Pointer;
  F: File;
  TF: TextFile;
  S: String;
  WS: WideString;
  StartTime: TDateTime;
  WC: Int64;
  StdOut: THandle;
  ConInfo: TConsoleScreenBufferInfo;
  StatsWritten: Boolean;
  Written: DWORD;
begin
  SetLength(EncryptedItems, 1000);
  if Options.EncryptedResourcesPrefix <> '' then
    for I := 0 to Length(EncryptedItems) - 1 do
    begin
      Res := FindResource(HInstance, PChar(Options.EncryptedResourcesPrefix + IntToStr(I)), RT_RCDATA);
      if Res = 0 then
      begin
        SetLength(EncryptedItems, I);
        Break;
      end;
      Resource := LoadResource(HInstance, Res);
      Win32Check(Resource <> 0);
      Data := LockResource(Resource);
      Win32Check(Data <> NIL);
      SetLength(EncryptedItems[I], SizeofResource(HInstance, Res));
      Move(Data^, EncryptedItems[I][1], Length(EncryptedItems[I]));
    end
  else if Options.EncryptedPath <> '' then
    for I := 0 to Length(EncryptedItems) - 1 do
    begin
      try
        AssignFile(F, Options.EncryptedPath + '\' + IntToStr(I));
        Reset(F, 1);
      except
        SetLength(EncryptedItems, I);
        Break;
      end;
      try
        SetLength(EncryptedItems[I], FileSize(F));
        BlockRead(F, EncryptedItems[I][1], Length(EncryptedItems[I]));
      finally
        CloseFile(F);
      end;
    end
  else
    SetLength(EncryptedItems, 0);

  SetLength(WorkKeys, 50000);

  if Options.KeyFile = '' then
    SetLength(WorkKeys, 0)
  else
  begin
    AssignFile(TF, Options.KeyFile);
    Reset(TF);
    try
      I := 1;
      KI := 0;
      while not EOF(TF) do
      begin
        ReadLn(TF, S);
        if (S <> '') and not (S[1] in [';', '#']) then
          try
            WorkKeys[KI].Key := ParseKeyLine(S);
            WorkKeys[KI].LineNumber := I;
            Inc(KI);
          except
            on E: Exception do
              WriteLn('Error on line #', I, ' in key file: ', E.Message);
          end;
        Inc(I);
      end;
      SetLength(WorkKeys, KI);
    finally
      CloseFile(TF);
    end;
  end;

  WorkMax := Int64(Length(WorkKeys)) * Length(WorkKeys) * Length(WorkKeys);

  if (WorkMax = 0) or (Length(EncryptedItems) = 0) then
    raise EAbort.Create('No lines. Nothing to do!');

  WriteLn(Format('Mining %d line + meme combinations using %d threads. There are %d keys for the win!', [Length(WorkKeys), Options.MaxThreads, Length(EncryptedItems)]));
  WriteLn;

  InitializeCriticalSection(WorkCounterCriticalSection);
  WorkCounter := 0;
  StartTime := Time;
  SetLength(Threads, Options.MaxThreads);

  for I := 0 to Options.MaxThreads - 1 do
  begin
    Threads[I] := BeginThread(NIL, 0, ThreadProc, NIL, 0, ID);
    if Threads[I] = 0 then
    begin
      SetLength(Threads, I);
      Break;
    end;
  end;

  if Length(Threads) = 0 then
    raise EAbort.Create('No threads could be created.');

  LastWon := 0;
  StdOut := GetStdHandle(STD_OUTPUT_HANDLE);
  StatsWritten := False;

  while WaitForMultipleObjects(Length(Threads), @Threads[0], True, 1000) = WAIT_TIMEOUT do
  begin
    I := Trunc(SecondSpan(Time, StartTime));
    if (I > 0) and GetConsoleScreenBufferInfo(StdOut, ConInfo) then
    begin
      if StatsWritten then
      begin
        ConInfo.dwCursorPosition.X := 0;
        Dec(ConInfo.dwCursorPosition.Y, 2);
        if not SetConsoleCursorPosition(StdOut, ConInfo.dwCursorPosition) then
          RaiseLastOSError;
        if not FillConsoleOutputCharacter(StdOut, ' ', ConInfo.dwSize.X * ConInfo.dwSize.Y, ConInfo.dwCursorPosition, Written) then
          RaiseLastOSError;
      end
      else
        StatsWritten := True;

      KI := InterlockedExchangeAdd(WonKeysCounter, 0);
      FlushWonKeys(KI);

      WC := WorkCounter;
      KI := Trunc((WorkMax - WC) / (WC / I));
      WriteLn(Format('%02.0f%%  %02d:%.2d %07.0d kh/s %s %02d:%.2d  %.12d/%.12d checked',
        [WC / WorkMax * 100, KI div 60, KI mod 60, Round(WC / I),
         StringRepeat(' ', ConInfo.dwSize.X - 75),
         MinutesBetween(Time, StartTime), SecondsBetween(Time, StartTime) mod 60,
         WC, WorkMax]));

      I := Trunc(WC / WorkMax * (ConInfo.dwSize.X - 2));
      Write(StringRepeat('=', I), '>');
      if KI > 365 * 24 * 3600 then
        Write('   Sorry, you''re not gonna live this long...');
      WriteLn;
    end;
  end;

  DeleteCriticalSection(WorkCounterCriticalSection);   // must be in try/finally but lazy today.

  if not FlushWonKeys(WonKeysCounter) then
    WriteLn;

  if WonKeysCounter = 0 then
    WriteLn('Finished. Nothing found :''(  Review your codez, enlarge your mad skillz!')
  else
  begin
    WriteLn(Format('Finished. You captured %d out of %d keys (I''ve also copied them to clipboard):', [WonKeysCounter, Length(EncryptedItems)]));
    WriteLn;
    WS := '';
    for I := 0 to WonKeysCounter - 1 do
    begin
      S := Copy(WonKeys[I].Decrypted, Length(WinMark) + 1, $FFFF);
      WriteLn('  ', S);
      WS := WS + S + #13#10;
    end;
    CopyToClipboard(WS);
  end;

  if Options.Wait then
  begin
    MessageBeep(MB_OK);
    WriteLn;
    WriteLn('Press Enter to close this window...');
    ReadLn;
  end;
end;

procedure PrintHelp;
begin
  ExitCode := 1;

  WriteLn('~True World~ Marble Cake Game');
  WriteLn;
  WriteLn('First argument must be a path to the text file with answers. If not,');
  WriteLn('you will be asked for it.');// and the result will be remembered for future');
//  WriteLn('invocation in an INI file in %TEMP%.');
  WriteLn;
  WriteLn('This thing is multi-threaded and will use all CPUs (if there are 1 or 2)');
  WriteLn('or N-1 CPUs (if more). Control this through %NUMBER_OF_PROCESSORS%.');
  WriteLn;
  WriteLn('To bypass the confirmation prompt on exit, use: echo | TrueCake.exe');
end;

procedure Run;
var
  Options: TMineOptions;
  S: String;
  OFN: TOpenFilename;
begin
  FillChar(Options, SizeOf(Options), 0);

  GetEnvironmentVariable('TEMP', @IniFile[0], SizeOf(IniFile) - 1 { ensure always ends on #0 });
  lstrcpyn(@IniFile[strlen(@IniFile[0])], '\TrueCake.ini', SizeOf(IniFile) - strlen(@IniFile[0]) - 1);

  if ParamStr(1) = '' then
  begin
    Options.KeyFile := ReadINI('General', 'DefaultCakeFile');
//    if not FileExists(Options.KeyFile) then
    begin
      FillChar(OFN, SizeOf(OFN), 0);
      OFN.lStructSize := SizeOf(OFN);
      SetLength(Options.KeyFile, MAX_PATH);
      OFN.lpstrFilter := 'A1L F!7e2'#0'*.*'#0;
      OFN.lpstrFile := @Options.KeyFile[1];
      OFN.nMaxFile := MAX_PATH;
      OFN.lpstrTitle := 'PiCK THe C@kE F!1e :^)';
      OFN.Flags := OFN_FILEMUSTEXIST + OFN_HIDEREADONLY;
      if GetOpenFileName(OFN) then
      begin
        SetLength(Options.KeyFile, strlen(@Options.KeyFile[1]));
        WriteINI('General', 'DefaultCakeFile', Options.KeyFile);
      end
      else
      begin
        PrintHelp;
        Exit;
      end
    end;
  end
  else if not FileExists(ParamStr(1)) then
  begin
    PrintHelp;
    Exit;
  end
  else
    Options.KeyFile := ParamStr(1);

{$IFDEF RELEASE}
  Options.EncryptedResourcesPrefix := 'loot';
{$ELSE}
  Options.EncryptedPath := 'C:\Encrypted-RU';
{$ENDIF}

  SetLength(S, 2);
  GetEnvironmentVariable('NUMBER_OF_PROCESSORS', @S[1], Length(S));
  Options.MaxThreads := Max(1, StrToInt(S));
  if Options.MaxThreads > 2 then
    Dec(Options.MaxThreads);

  Options.Wait := True;
//Options.MaxThreads := 1;

  MineForTheWin(Options);
end;

begin
  try
    // FastMM mistakes a global string (like Options.KeyFile) for leaked memory
    // since it isn't deallocated when FastMM shuts down.
    Run;
  except
    on E: Exception do
    begin
      WriteLn('Error: ', E.Message);
      ExitCode := 1;
    end;
  end;
end.
