unit UnicodeUtils;

interface

uses Windows;

function ExtractFilePath(FileName: WideString): WideString;
function ExtractFileName(Path: WideString): WideString;    
function ExtractFileExt(FileName: WideString): WideString;   
function ChangeFileExt(FileName, Extension: WideString): WideString;
function BaseName(Path: WideString): WideString;
function ChDir(const ToPath: WideString): Boolean;

function IncludeTrailingPathDelimiter(Path: WideString): WideString;
function ExcludeTrailingPathDelimiter(Path: WideString): WideString;
function FileInfo(Path: WideString): TWin32FindDataW;
function IsDirectory(Path: WideString): Boolean;
function FileExists(Path: WideString): Boolean;

function ExpandFileName(FileName: WideString): WideString;    
function ParamStrW(Index: Integer): WideString;

implementation

function ExtractFilePath(FileName: WideString): WideString;
var
  I: Word;
begin
  Result := '';
  I := Length(FileName);
  while (I >= 1) and (FileName[I] <> '\') and (FileName[I] <> ':') and (FileName[I] <> '/') do
    Dec(I);
  Result := Copy(FileName, 1, I);
end;

function ExtractFileName(Path: WideString): WideString;
var
  I: Word;
begin
  for I := Length(Path) downto 1 do
    if (Path[I] = '\') or (Path[I] = ':') or (Path[I] = '/') then
    begin
      Result := Copy(Path, I + 1, Length(Path));
      Exit
    end;

  Result := Path
end;

function BaseName(Path: WideString): WideString;
begin
  Result := ExtractFileName(Path);
end;

function ChDir(const ToPath: WideString): Boolean;
begin
  Result := SetCurrentDirectoryW(PWideChar(ToPath));
end;

function IncludeTrailingPathDelimiter(Path: WideString): WideString;
begin
  Result := Path;
  if (Result = '') or ((Result[Length(Result)] <> '\') and (Result[Length(Result)] <> '/')) then
    Insert('\', Result, $FFFF);
end;

function ExcludeTrailingPathDelimiter(Path: WideString): WideString;
begin
  Result := Path;
  if (Result <> '') and ((Result[Length(Result)] = '\') or (Result[Length(Result)] = '/')) then
    Delete(Result, Length(Result), 1);
end;

function FileInfo(Path: WideString): TWin32FindDataW;
var
  Handle: DWord;
begin
  Handle := FindFirstFileW(PWideChar( ExcludeTrailingPathDelimiter(Path) ), Result);
  if Handle <> INVALID_HANDLE_VALUE then
    Windows.FindClose(Handle)
    else
      Result.ftLastWriteTime.dwLowDateTime := 0
end;

function IsDirectory(Path: WideString): Boolean;
begin
  with FileInfo(Path) do
    Result := (ftLastWriteTime.dwLowDateTime <> 0) and (dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY <> 0)
end;

function FileExists(Path: WideString): Boolean;
begin
  with FileInfo(Path) do
    Result := (ftLastWriteTime.dwLowDateTime <> 0) and (dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0)
end;

function ExpandFileName(FileName: WideString): WideString;
var
  Name: PWideChar;
begin
  SetLength(Result, $1000);
  SetLength(Result, GetFullPathNameW(PWideChar(FileName), $1000, PWideChar(Result), Name));
end;

function ExtractFileExt(FileName: WideString): WideString;
var
  I: Word;
begin
  for I := Length(FileName) downto 1 do
    if FileName[I] = '\' then
      Break
      else if FileName[I] = '.' then
      begin
        Result := Copy(FileName, I, Length(FileName));
        Exit
      end;

  Result := ''
end;

function ChangeFileExt(FileName, Extension: WideString): WideString;
var
  I: Word;
begin
  for I := Length(FileName) downto 1 do
    if FileName[I] = '\' then
      Break
      else if FileName[I] = '.' then
      begin
        Result := Copy(FileName, 1, I - 1) + Extension;
        Exit
      end;

  Result := FileName + Extension;
end;

function ParamStrW(Index: Integer): WideString;
var
  I, CurrentIndex: Word;
  CmdLine: WideString;
  Join: Boolean;
begin
  if Index = 0 then
  begin
    SetLength(Result, MAX_PATH);
    SetLength(Result, GetModuleFileNameW(0, @Result[1], MAX_PATH));
    Exit
  end;

  CmdLine := GetCommandLineW;

  Result := '';
  Join := False;
  CurrentIndex := 0;

  for I := 1 to Length(CmdLine) do
    if CmdLine[I] = '"' then
      Join := not Join
      else if (CmdLine[I] = ' ') and not Join then
        Inc(CurrentIndex)
        else if CurrentIndex = Index then
          Result := Result + CmdLine[I]
          else if CurrentIndex > Index then
            Exit
end;

end.
